module MmioRegister (
    input clk,
    input reset,

    input [31:0] DataAddr,
    
    input [31:0] WriteData,
    input [3:0] WriteMask,
    input MemWrite,
    
    output [31:0] ReadData,

    input [31:0] d,
    output reg [31:0] q,
    output selector
);
    parameter MMIO_ADDRESS = 32'h80000000;


    assign selector = DataAddr == MMIO_ADDRESS;

    always @(posedge clk or posedge reset) begin
        if (reset)
            q <= 32'd0;
        else
            if (selector & MemWrite) begin
                if(WriteMask[0])
                    q[7:0] <= WriteData[7:0];
                if(WriteMask[1])
                    q[15:8] <= WriteData[15:8];
                if(WriteMask[2])
                    q[23:16] <= WriteData[23:16];
                if(WriteMask[3])
                    q[31:24] <= WriteData[31:24];
            end
    end

    tbuf32 read_data_buffer (
        .d(d),
        .q(ReadData),
        .en(selector)
    );
endmodule


module Mmio (
    input clk,
    input reset,

    input [31:0] DataAddr,
    
    input [31:0] WriteData,
    input [3:0] WriteMask,
    input MemWrite,
    
    output [31:0] ReadData,


    inout [31:0] gpio_pins,
    output tx
);

    GPIO gpio (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .gpio_pins(gpio_pins)
    );
    defparam gpio.MMIO_ADDRESS =  32'h80000000;

    UART uart (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .tx(tx)
    );
    defparam uart.MMIO_ADDRESS =  32'h80000008;

endmodule


module GPIO (
    input clk,
    input reset,

    input [31:0] DataAddr,
    
    input [31:0] WriteData,
    input [3:0] WriteMask,
    input MemWrite,
    
    output [31:0] ReadData,

    inout [31:0] gpio_pins
);
    parameter MMIO_ADDRESS = 32'h80000000;
    
    wire [31:0] mode;
    MmioRegister mode_register (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .d(mode),
        .q(mode)
    );
    defparam mode_register.MMIO_ADDRESS = MMIO_ADDRESS;

    wire [31:0] gpio_out;
    MmioRegister gpio_register (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .d(gpio_pins),
        .q(gpio_out)
    );
    defparam gpio_register.MMIO_ADDRESS = MMIO_ADDRESS + 4;

    genvar i;
    generate
        for (i = 0; i < 32; i = i + 1) begin
            assign gpio_pins[i] = mode[i] ? gpio_out[i] : 1'bz;
        end
    endgenerate
endmodule



module UART (
    input clk,
    input reset,

    input [31:0] DataAddr,
    
    input [31:0] WriteData,
    input [3:0] WriteMask,
    input MemWrite,
    
    output [31:0] ReadData,

    output tx
);
    parameter MMIO_ADDRESS = 32'h80000000;

    wire [31:0] scaler;
    MmioRegister scaler_register (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .d(scaler),
        .q(scaler)
    );
    defparam scaler_register.MMIO_ADDRESS = MMIO_ADDRESS;

    UARTTX uart_tx (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .scaler(scaler),
        .tx(tx)
    );
    defparam uart_tx.MMIO_ADDRESS = MMIO_ADDRESS + 4;
endmodule

module UARTTX (
    input clk,
    input reset,

    input [31:0] DataAddr,
    
    input [31:0] WriteData,
    input [3:0] WriteMask,
    input MemWrite,
    
    output [31:0] ReadData,

    input [31:0] scaler,

    output reg tx
);
    parameter MMIO_ADDRESS = 32'h80000000;


    localparam STATE_IDLE	= 2'b00;
    localparam STATE_START	= 2'b01;
    localparam STATE_DATA	= 2'b10;
    localparam STATE_STOP	= 2'b11;

    wire [31:0] data;
    reg [2:0] bitpos = 3'h0;
    reg [1:0] state;

    wire selector;
    MmioRegister register (
        .clk(clk),
        .reset(reset),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .WriteMask(WriteMask),
        .MemWrite(MemWrite),
        .ReadData(ReadData),
        .d({30'd0, state}),
        .q(data),
        .selector(selector)
    );
    defparam register.MMIO_ADDRESS = MMIO_ADDRESS;


    reg[32:0] tx_clkcnt;
    wire tx_clk = (tx_clkcnt == 0);

    always @(posedge clk or posedge reset) begin
        if (reset) begin
            tx_clkcnt <= 0;
        end else begin
            if (tx_clkcnt >= scaler) begin
                tx_clkcnt <= 0;
            end else begin
                tx_clkcnt <= tx_clkcnt + 1;
            end
        end
    end

    always @(posedge clk or posedge reset) begin
        if (reset) begin
            tx <= 1'b1;
            state <= STATE_IDLE;
        end else begin
            case (state)
                STATE_IDLE:
                    begin
                        if (MemWrite & selector) begin
                            state <= STATE_START;
                            bitpos <= 3'h0;
                        end
                    end
                STATE_START:
                    begin
                        if (tx_clk) begin
                            tx <= 1'b0;
                            state <= STATE_DATA;
                        end
                    end
                STATE_DATA:
                    begin
                        if (tx_clk) begin
                            if (bitpos == 3'h7)
                                state <= STATE_STOP;
                            else
                                bitpos <= bitpos + 3'h1;
                            tx <= data[bitpos];
                        end
                    end
                STATE_STOP:
                    begin
                        if (tx_clk) begin
                            tx <= 1'b1;
                            state <= STATE_IDLE;
                        end
                    end
            endcase
        end
    end
endmodule