module RiscV (
    input clk,
    input reset,
    output [31:0] PC,
    input [31:0] Instr,
    output MemWrite,
    output [31:0] DataAddr,
    output [31:0] WriteData,
    input [31:0] ReadData,
    output [3:0] WriteMask
);
    wire [1:0] ResultSrc;
    wire PCSrc;
    wire AluSrc;
    wire RegWrite;
    wire [2:0] ImmSrc;
    wire [1:0] AluCtl;
    wire PCAluResult;

    wire JumpPredicate;


    Controller controller (
        .op(Instr[6:0]),
        .funct3(Instr[14:12]),
        .funct7b5(Instr[30]),
        .JumpPredicate(JumpPredicate),

        .ResultSrc(ResultSrc),
        .MemWrite(MemWrite),
        .PCSrc(PCSrc),
        .AluSrc(AluSrc),
        .RegWrite(RegWrite),
        .ImmSrc(ImmSrc),
        .AluCtl(AluCtl),
        .PCAluResult(PCAluResult)
    );

    ExecutionUnit execution_enit (
        .clk(clk),
        .reset(reset),

        .ResultSrc(ResultSrc),
        .PCSrc(PCSrc),
        .AluSrc(AluSrc),
        .RegWrite(RegWrite),
        .ImmSrc(ImmSrc),
        .AluCtl(AluCtl),
        .PCAluResult(PCAluResult),
        .Instr(Instr),
        .JumpPredicate(JumpPredicate),
        .PC(PC),
        .DataAddr(DataAddr),
        .WriteData(WriteData),
        .ReadData(ReadData),
        .WriteMask(WriteMask)
    );

endmodule

module Controller (
    input [6:0] op,
    input [2:0] funct3,
    input funct7b5,
    input JumpPredicate,

    output [1:0] ResultSrc,
    output MemWrite,
    output PCSrc,
    output AluSrc,
    output RegWrite,
    output [2:0] ImmSrc,
    output [1:0] AluCtl,
    output PCAluResult
);

    reg [12:0] controlls;

    wire Jump;
    wire Branch;

    assign {PCAluResult, RegWrite, ImmSrc, AluSrc, MemWrite, ResultSrc, Branch, Jump, AluCtl} = controlls;

    always @(*) begin
        case (op)
            7'b0010011: controlls <= 13'b0_1_000_1_0_00_0_0_00; // Integer Register-Immediate Instructions
            7'b0110011: controlls <= 13'b0_1_000_0_0_00_0_0_00; // Integer Register-Register Operations
            7'b1101111: controlls <= 13'b0_1_011_0_0_10_0_1_00; // Unconditional Jumps (jal)
            7'b1100111: controlls <= 13'b1_1_000_1_0_10_0_1_01; // Unconditional Jumps (jalr)
            7'b1100011: controlls <= 13'b0_0_010_0_0_00_1_0_00; // Conditional branches
            7'b0000011: controlls <= 13'b0_1_000_1_0_01_0_0_01; // Load Instructions
            7'b0100011: controlls <= 13'b0_0_001_1_1_00_0_0_01; // Store Instructions
            
            7'b0110111: controlls <= 13'b0_1_100_0_0_11_0_0_00; // lui
            7'b0010111: controlls <= 13'b0_1_100_1_0_00_0_0_10; // auipc            
            
            default: controlls <= 13'bx_x_xxx_x_x_xx_x_x_xx;
        endcase 
    end

    assign PCSrc = Branch & JumpPredicate | Jump;
endmodule

module ExecutionUnit (
    input clk,
    input reset,

    input [1:0] ResultSrc,
    input PCSrc,
    input AluSrc,
    input RegWrite,
    input [2:0] ImmSrc,
    input [1:0] AluCtl,
    input PCAluResult,

    input [31:0] Instr,

    output JumpPredicate,

    output [31:0] PC,

    output [31:0] DataAddr,
    output [31:0] WriteData,
    input [31:0] ReadData,
    output [3:0] WriteMask
);
    wire [31:0] PCNext;
    wire [31:0] PCPlus4;
    wire [31:0] PCTarget;
    wire [31:0] ImmExt;
    wire [31:0] SrcA;
    wire [31:0] SrcB;
    wire [31:0] Result;
    wire [31:0] AluResult;
    wire [31:0] MaskedReadData;
    assign DataAddr = AluResult;

    reg32 pc_reg (
        .clk(clk),
        .reset(reset),
        .d(PCNext),
        .q(PC)
    );

    assign PCPlus4 = PC + 32'd4;
    assign PCTarget = PC + ImmExt;


    MaskUnit mask_unit (
        .funct3(Instr[14:12]),
        .DataAddr(DataAddr),
        .WriteMask(WriteMask),
        .ReadData(ReadData),
        .MaskedReadData(MaskedReadData)
    );

    wire [31:0] pctmp;
    mux32_2 pc_mux (
        .a(PCPlus4),
        .b(PCTarget),
        .sel(PCSrc),
        .out(pctmp)
    );

    mux32_2 pc_alu_mux (
        .a(pctmp),
        .b(AluResult),
        .sel(PCAluResult),
        .out(PCNext)
    );

    initial begin
        // $monitor("PC: %x", PC);    
    end

    RegisterFile register (
        .clk(clk),

        .we3(RegWrite),
        .wd3(Result),

        .a1(Instr[19:15]),
        .a2(Instr[24:20]),
        .a3(Instr[11:7]),
        .rd1(SrcA),
        .rd2(WriteData)
    );

    ImmExtend imm_extenf (
        .Instr(Instr[31:7]),
        .ImmSrc(ImmSrc),
        .ImmExt(ImmExt)
    );

    mux32_2 src_mux (
        .a(WriteData),
        .b(ImmExt),
        .sel(AluSrc),
        .out(SrcB)
    );

    Alu alu (
        .SrcA(SrcA),
        .SrcB(SrcB),
        .PC(PC),
        .AluResult(AluResult),
        .JumpPredicate(JumpPredicate),
        .funct3(Instr[14:12]),
        .funct7b5(Instr[30]),
        .opb5(Instr[5]),
        .AluCtl(AluCtl)
    );

    mux32_4 result_mux (
        .a(AluResult),
        .b(MaskedReadData),
        .c(PCPlus4),
        .d(ImmExt),
        .sel(ResultSrc),
        .out(Result)
    );
endmodule

module MaskUnit (
    input [2:0] funct3,
    input [31:0] DataAddr,
    output reg [3:0] WriteMask,

    input [31:0] ReadData,
    output reg [31:0] MaskedReadData
);

    wire isByte;
    assign isByte = funct3[1:0] == 2'b00;
    wire isHalf;
    assign isHalf = funct3[1:0] == 2'b01;



    wire [15:0] half;
    assign half = DataAddr[1] ? ReadData[31:16] : ReadData[15:0];
    wire [7:0] byte;
    assign byte = DataAddr[0] ? half[15:8] : half[7:0];


    always @(*) begin
        if (isByte)
            MaskedReadData <= {{24{sign}}, byte};
        else if (isHalf)
            MaskedReadData <= {{16{sign}}, half};
        else
            MaskedReadData <= ReadData;
    end

    always @(*) begin
        if (isByte)
            case (DataAddr[1:0])
                2'b00: WriteMask <= 4'b0001;
                2'b01: WriteMask <= 4'b0010;
                2'b10: WriteMask <= 4'b0100;
                2'b11: WriteMask <= 4'b1000;
            endcase
        else if (isHalf)
            if (DataAddr[1])
                WriteMask <= 4'b1100;
            else
                WriteMask <= 4'b0011;
        else
            WriteMask <= 4'b1111;
    end


    wire sign;
    assign sign = !funct3[2] & (isByte ? byte[7] : half[15]);

endmodule

module Alu (
    input [31:0] SrcA,
    input [31:0] SrcB,
    input [31:0] PC,
    output reg [31:0] AluResult,
    output reg JumpPredicate,
    input [2:0] funct3,
    input funct7b5,
    input opb5,
    input [1:0] AluCtl
);

    wire [31:0] plus = SrcA + SrcB;
    wire [32:0] minus = {1'b1, ~SrcB} + {1'b0, SrcA} + 33'b1;

    wire lt = (SrcA[31] ^ SrcB[31]) ? SrcA[31] : minus[32];
    wire ltu = minus[32];
    wire eq = (minus[31:0] == 0);

    wire signed [31:0] SrcASigned;
    assign SrcASigned = SrcA;
    wire [31:0] ArithmeticShift = SrcASigned >>> SrcB[4:0];

    always @(*) begin
        case (AluCtl)
            2'b00:
                case (funct3)
                    3'b000: AluResult <= (funct7b5 & opb5) ? minus : plus;
                    3'b001: AluResult <= SrcA << SrcB[4:0];
                    3'b010: AluResult <= lt;
                    3'b011: AluResult <= ltu;
                    3'b100: AluResult <= SrcA ^ SrcB;
                    3'b101: AluResult <= (funct7b5) ? (ArithmeticShift) : (SrcA >> SrcB[4:0]);
                    3'b110: AluResult <= SrcA | SrcB;
                    3'b111: AluResult <= SrcA & SrcB;
                    default: AluResult <= 32'bxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx;
                endcase
            2'b01: AluResult <= plus;
            2'b10: AluResult <= PC + SrcB;
            // 2'b11: 
            default: AluResult <= 32'bxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx;
        endcase
    end

    always @(*) begin
        case (funct3)
            3'b000: JumpPredicate <= eq;
            3'b001: JumpPredicate <= !eq;
            3'b100: JumpPredicate <= lt;
            3'b101: JumpPredicate <= !lt;
            3'b110: JumpPredicate <= ltu;
            3'b111: JumpPredicate <= !ltu;
            default: JumpPredicate <= 1'bx;
        endcase
    end
endmodule

module RegisterFile (
    input clk,
    
    input [4:0] a1,
    input [4:0] a2,
    input [4:0] a3,

    input we3,
    input [31:0] wd3,

    output [31:0] rd1,
    output [31:0] rd2
);
    reg [31:0] rf [31:0];

    always @(posedge clk) begin
        if (we3) begin
            rf[a3] <= wd3;
            // $display("x%d <- %x", a3, wd3);            
        end
    end

    assign rd1 = (a1 != 0) ? rf[a1] : 32'd0;
    assign rd2 = (a2 != 0) ? rf[a2] : 32'd0;
endmodule


module ImmExtend (
    input [31:7] Instr,
    input [2:0] ImmSrc,
    output reg [31:0] ImmExt
);
    always @(*) begin
        case (ImmSrc)
            3'b000: ImmExt <= {{20{Instr[31]}}, Instr[31:20]};
            3'b001: ImmExt <= {{20{Instr[31]}}, Instr[31:25], Instr[11:7]};
            3'b010: ImmExt <= {{20{Instr[31]}}, Instr[7], Instr[30:25], Instr[11:8], 1'b0};
            3'b011: ImmExt <= {{12{Instr[31]}}, Instr[19:12], Instr[20], Instr[30:21], 1'b0};
            3'b100: ImmExt <= {Instr[31:12], 12'd0};
            default: ImmExt <= 32'bxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx;
        endcase
    end
endmodule
